"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const allowedKeys = [
    'tagName',
    'metadata',
    'flow',
    'sectioning',
    'heading',
    'phrasing',
    'embedded',
    'interactive',
    'deprecated',
    'void',
    'transparent',
    'implicitClosed',
    'deprecatedAttributes',
    'permittedContent',
    'permittedDescendants',
    'permittedOrder',
];
const dynamicKeys = [
    'metadata',
    'flow',
    'sectioning',
    'heading',
    'phrasing',
    'embedded',
    'interactive',
];
const functionTable = {
    isDescendant,
    hasAttribute,
    matchAttribute,
};
class MetaTable {
    constructor() {
        this.elements = {};
    }
    loadFromObject(obj) {
        for (const key of Object.keys(obj)) {
            this.addEntry(key, obj[key]);
        }
    }
    loadFromFile(filename) {
        this.loadFromObject(require(filename));
    }
    getMetaFor(tagName) {
        tagName = tagName.toLowerCase();
        return this.elements[tagName] ? Object.assign({}, this.elements[tagName]) : null;
    }
    addEntry(tagName, entry) {
        for (const key of Object.keys(entry)) {
            if (allowedKeys.indexOf(key) === -1) {
                throw new Error(`Metadata for <${tagName}> contains unknown property "${key}"`);
            }
        }
        this.elements[tagName] = Object.assign({
            tagName,
            void: false,
        }, entry);
    }
    resolve(node) {
        if (node.meta) {
            expandProperties(node, node.meta);
        }
    }
}
exports.MetaTable = MetaTable;
function expandProperties(node, entry) {
    for (const key of dynamicKeys) {
        const property = entry[key];
        if (property && typeof property !== 'boolean') {
            entry[key] = evaluateProperty(node, property);
        }
    }
}
function evaluateProperty(node, expr) {
    const [func, options] = parseExpression(expr);
    return func(node, options);
}
function parseExpression(expr) {
    if (typeof expr === 'string') {
        return parseExpression([expr, {}]);
    }
    else {
        const [funcName, options] = expr;
        const func = functionTable[funcName];
        if (!func) {
            throw new Error(`Failed to find function when evaluation property expression "${expr}"`);
        }
        return [func, options];
    }
}
function isDescendant(node, tagName) {
    if (typeof tagName !== 'string') {
        throw new Error(`Property expression "isDescendant" must take string argument`);
    }
    let cur = node.parent;
    while (!cur.isRootElement()) {
        if (cur.is(tagName)) {
            return true;
        }
        cur = cur.parent;
    }
    return false;
}
function hasAttribute(node, attr) {
    if (typeof attr !== 'string') {
        throw new Error(`Property expression "hasAttribute" must take string argument`);
    }
    return node.hasAttribute(attr);
}
function matchAttribute(node, match) {
    if (!Array.isArray(match) || match.length !== 3) {
        throw new Error(`Property expression "matchAttribute" must take [key, op, value] array as argument`);
    }
    const [key, op, value] = match.map(x => x.toLowerCase());
    const nodeValue = (node.getAttributeValue(key) || '').toLowerCase();
    switch (op) {
        case '!=': return nodeValue !== value;
        case '=': return nodeValue === value;
        default: throw new Error(`Property expression "matchAttribute" has invalid operator "${op}"`);
    }
}
